(in-package "WEYLI")

(defclass ge-integral (general-expression)
  ((expr :initarg :expr
	 :accessor expression-of)
   (var :initarg :var
	:accessor variable-of)
   (lower :initarg :lower
	  :initform nil
	  :accessor lower-bound-of)
   (upper :initarg :upper
	  :initform nil
	  :accessor upper-bound-of)))

(defmethod print-object ((int ge-integral) stream)
  (format stream "int ~A d~A" (expression-of int) (variable-of int))
  (if (lower-bound-of int)
      (format stream " from ~A to ~A" (lower-bound-of int)
	      (upper-bound-of int))))

(defmethod integral (expr var)
  (make-instance 'ge-integral :expr expr :var var :domain *general*))

(defmethod integral ((expr domain-element) var)
  (make-instance 'ge-integral :expr expr :var var :domain (domain-of expr)))

(defun definite-integral (expr var lower upper)
  (let* ((ge-var (coerce var *general*))
	 (indef (integral expr ge-var)))
    (if (typep indef 'ge-integral)
	(make-instance 'ge-integral :expr expr :var var
		       :domain (domain-of indef)
		       :lower lower :upper upper)
      (- (substitute (coerce upper *general*) ge-var indef)
	 (substitute (coerce lower *general*) ge-var indef)))))

(defun constant-of-integration ()
  (coerce (gensym "C") *general*))

(defun indefinite-integral (expr var)
  (+ (integral expr (coerce var *general*))
     (constant-of-integration)))

(defun first-domain (&rest args)
  (cond ((null args) nil)
	((typep (car args) 'domain-element)
	 (domain-of (car args)))
	(t (apply #'first-domain (cdr args)))))

(defun int (expr var &optional lower upper)
  (if lower
      (weyli::definite-integral expr var lower upper)
    (weyli::indefinite-integral expr var)))

(defmethod symbol-of ((sym symbol))
  sym)

(export 'int)

;;; useful specializations follow

(defmethod integral ((expr (or number numeric)) var)
  (* expr var))

;; (defmethod integral ((expr (or symbol ge-variable)) var)
;;   (let ((domain (or (first-domain expr var) *general*)))
;;     (if (= (coerce expr domain) (coerce var domain))
;; 	   (* 0.5 var var)
;; 	 (* expr var))))  

(defmethod integral ((expr (or symbol ge-variable)) var)
  (if (apply #'eq (mapcar #'symbol-of (list expr var)))
      (* 1/2 var var)
    ;; assumes independence of variables
    (* expr var)))

(defmethod integral ((expr ge-plus) var)
  (make-ge-plus (domain-of expr)
		(mapcar (lambda (exp)
			  (integral exp var))
			(terms-of expr))))

;; Can't specialize on multiplication by a constant without predicate
;; dispatch (or an inference system, which would probably help much more
;; for the general problem of integration.)

;; the multivariate-polynomial code is a mess.
;;(defmethod integral ((expr mpolynomial) var)
;;  ...)

;;(defmethod integral ((expr upolynomial) var)
;;  ...)

(defmethod integral ((expr ge-application) var)
  (let ((domain (domain-of expr))
	(fname (name-of (funct-of expr))))
    (cond
     ((equal fname "sin") (- (make-ge-cos domain (first (args-of expr)))))
     ((equal fname "cos") (make-ge-sin domain (first (args-of expr))))
     ;;...
     (t (call-next-method)))))
