(require 'pd-package)
(in-package :predicate-dispatch)

(require 'predicate-classes)
(require 'xcond)

;; I'd love to use macroexpand-all from the walker bundled with PCL,
;; but it's not completely safe.  (Tagbody's tags *cannot* result from
;; macro expansion.)

(defun build-predicate (lambda-list predicate-bodies)
  (multiple-value-bind (ll-analysis ll-predicates auxv-count)
      (analyze-lambda-list-for-analysis lambda-list)
    (make-instance 'predicate-qualifier
		   :predicate (normalize-predicate
			       (make-and
				(append ll-predicates
					(mapcar #'(lambda (body)
						    (build-pred-internal
						     (analyze-expr body
								   ll-analysis)
						     lambda-list))
						predicate-bodies))))
		   :auxv-count auxv-count)))

(defclass expression-analysis ()
  ((arguments-used :initarg :args-used
		   :reader args-used)))

(defclass trivial-analysis (expression-analysis)
  ((code :initarg :code
	 :reader code-of)
   (substitute? :initarg :substitute?
		:reader substitute?
		:initform nil)))

(defclass manifest-constant (expression-analysis)
  ((value :initarg :value
	  :reader value-of)
   (arguments-used :initform nil)))

(defclass extraction-or-simple-test (expression-analysis)
  ((argument-index :initarg :index
		   :reader index-of)
   (unary-chain :initarg :chain
		:reader chain-of)))

(defclass extraction-analysis (extraction-or-simple-test)
  ())

(defclass typecheck-analysis (extraction-or-simple-test)
  ((target-type :initarg :target
		:reader target-of)))

(defclass eql-analysis (extraction-or-simple-test)
  ((target-value :initarg :target
		 :reader target-of)))

(defclass not-analysis (expression-analysis)
  ((base-analysis :initarg :base
		  :reader base-of)))

(defclass compound-analysis (expression-analysis)
  ((terms :initarg :terms
	  :reader terms-of)))

(defclass and-analysis (compound-analysis)
  ())

(defclass or-analysis (compound-analysis)
  ())

(defun mkxtriv (analysis var code)
  (make-instance 'trivial-analysis
		 :args-used (adjoin var (args-used analysis))
		 :code code))

(defun mktriventry (arg)
  (cons arg (make-instance 'trivial-analysis :args-used (list arg) :code arg)))

(defun null-triv (expr)
  (make-instance 'trivial-analysis :args-used nil :code expr))

(defun car-or-identity (x)
  (if (consp x)
      (car x)
    x))

(defun extract-arg-name (ll-term key?)
  (cond ((atom ll-term) ll-term)
	((and (consp (car ll-term)) key?) (second ll-term))
	(t (car ll-term))))

(defun analyze-lambda-list-for-analysis (lambda-list)
  (do ((ll lambda-list (cdr ll))
       (index 0 (1+ index))
       (partial-result (list (mktriventry 'auxv)))
       (ll-predicates nil)
       (auxv-count 0)
       (last-key nil))
      ((null ll) (values partial-result (reverse ll-predicates) auxv-count))
    (let* ((carll (car ll))
	   (arg (extract-arg-name carll (eq last-key '&key))))
      (assert (symbolp arg))
      (cond ((eq carll '&allow-other-keys) nil) ; do nothing
	    ((member carll '(&optional &rest &key &aux)) (setf last-key carll))
	    ((member carll lambda-list-keywords)
	     (error "Unsupported lambda list keyword ~S" carll))
	    ((null last-key)
	     (push (cons arg (make-instance 'extraction-analysis
					    :index index :chain nil
					    :args-used (list arg)))
		   partial-result)
	     (if (and (consp carll) (cdr carll))
		 (push (build-pred-internal
			(analyze-specializer (second carll) arg index
					     partial-result)
			lambda-list)
		       ll-predicates)))
	    ((eq last-key '&rest) (push (mktriventry arg) partial-result))
	    ((eq last-key '&aux)
	     (if (consp carll)
		 (push (build-pred-internal
			(analyze-expr `(progn (setf (aref auxv ,auxv-count)
						    ,(second carll))
					      t)
				      partial-result)
			lambda-list)
		       ll-predicates))
	     (push (cons arg (make-instance 'trivial-analysis
					    :args-used `(auxv ,arg)
					    :code `(aref auxv ,auxv-count)
					    :substitute? t))
		   partial-result)
	     (incf auxv-count))
	    (t ; &optional or &key
	     (if (or (atom carll) (null (cdr carll)))
		 (push (mktriventry arg) partial-result)
	       (let ((analysis (analyze-expr (second carll) partial-result)))
		 (push (cons arg (mkxtriv analysis arg
					  `(or ,arg ,(second carll))))
		       partial-result)
		 (if (cddr carll)
		     (push (mktriventry (third carll)) partial-result)))))))))

(defun analyze-specializer (specializer arg index ll-analysis)
  (cond ((symbolp specializer) (make-instance 'typecheck-analysis
					      :index index :chain nil
					      :target specializer
					      :args-used (list arg)))
	((and (listp specializer)
	      (= (length specializer) 2)
	      (eq (car specializer) 'eql))
	 (analyze-expr `(eql ,arg ,(cadr specializer)) ll-analysis))
	(t (error "Unsupported specializer ~S for ~S" specializer arg))))

(defun analyze-expr (expr ll-analysis &optional macros symbol-macros)
  (multiple-value-bind (expansion expanded?)
      (macroexpand-1 expr)
    (xcond ((assoc expr symbol-macros)
	    => #'(lambda (a) (analyze-expr (cdr a) ll-analysis macros
					   symbol-macros)))
	   ((and (consp expr)
		 (assoc (car expr) macros))
	    => #'(lambda (a) (analyze-expr (apply (cdr a) (cdr expr))
					   ll-analysis macros symbol-macros)))
	   ((and expanded? (or (atom expr)
			       (not (or (special-operator-p (car expr))
					(member (car expr)
						'(and eql not or typep))))))
	    (analyze-expr expansion ll-analysis macros symbol-macros))
	   ((assoc expr ll-analysis) => #'cdr)
	   ((constantp expr) (make-instance 'manifest-constant
					    :value (eval expr)))
	   ((atom expr) (null-triv expr))
	   ((consp (car expr));; must be a lambda form
	    (analyze-exprs expr ll-analysis macros symbol-macros))
	   ((assoc (car expr) *analysis-helpers*)
	    => #'(lambda (a) (funcall (cdr a) expr ll-analysis macros
				      symbol-macros)))
	   (t (let* ((aa (mapcar #'(lambda (x) (analyze-expr x ll-analysis
							     macros
							     symbol-macros))
				 (cdr expr)))
		     (aa1 (car aa)))
		(if (or (cdr aa)
			(special-operator-p (car expr))
			(not (typep aa1 'extraction-analysis)))
		    (make-instance 'trivial-analysis :code expr
				   :args-used
				   (reduce #'union aa :key #'args-used))
		  (make-instance 'extraction-analysis
				 :index (index-of aa1)
				 :chain (cons (symbol-function (car expr))
					      (chain-of aa1))
				 :args-used (args-used aa1))))))))

(defun analyze-exprs (ee l m s &optional whole (force-trivial? t))
  (if (or force-trivial? (cdr ee))
      (make-instance 'trivial-analysis
		     :args-used (reduce #'union ee
					:key #'(lambda (e)
						 (args-used (analyze-expr
							     e l m s))))
		     :code whole)
    (analyze-expr (car ee) l m s)))

;; force trivial analysis
(defun analyze-first-arg (e l m s)
  (analyze-exprs (list (second e)) l m s e))

(defun analyze-second-arg (e l m s)
  (analyze-exprs (list (third e)) l m s e))

(defun analyze-block (e l m s)
  (analyze-exprs (cddr e) l m s e))

(defun analyze-eval-when (e l m s)
  (analyze-exprs (cddr e) l m s e))

(defun analyze-flet/labels (e l m s)
  (let* ((function-names (mapcar #'car (second e)))
	 (m2 (remove-if #'(lambda (x) (member (car x) function-names)) m))
	 (m-for-functions (case (car e)
			    (flet m)
			    (labels m2)))
	 (functions-analysis (analyze-exprs (mapcar #'(lambda (x)
							`(lambda ,@(cdr x)))
						    (second e))
					    l m-for-functions s))
	 (body-analysis (analyze-locally `(locally ,@(cddr e)) l m2 s)))
    (make-instance 'trivial-analysis
		   :args-used (union (args-used functions-analysis)
				     (args-used body-analysis))
		   :code e)))

(defun analyze-function (e l m s)
  (if (and (consp (cadr e))
	   (eq (caadr e) 'lambda))
      (analyze-lambda (cadr e) l m s)
    (null-triv e)))

(defun analyze-lambda (e l m s)
  (let ((new-bound-vars (mapcar #'car (analyze-lambda-list-for-analysis
				       (second e)))))
    (flet ((shadowed? (a) (member (car a) new-bound-vars)))
      (analyze-locally (if (stringp (third e))
			   (cdddr e)
			 (cddr e))
		       (remove-if #'shadowed? l)
		       m
		       (remove-if #'shadowed? s)
		       e))))

(defun analyze-let/let* (e l m s)
  (let ((new-bound-vars (mapcar #'car-or-identity (second e))))
    (flet ((shadowed? (a) (member (car a) new-bound-vars))
	   (for-bindings (x x*) (case (car e)
				  (let x)
				  (let* x*))))
      (let* ((l2 (remove-if #'shadowed? l))
	     (s2 (remove-if #'shadowed? s))
	     (bindings-analysis (analyze-exprs
				 (apply #'append
					(mapcar #'cdr
						(remove-if #'atom (second e))))
				 (for-bindings l l2) m (for-bindings s s2)))
	     (body-analysis (analyze-locally `(locally ,@(cddr e)) l2 m s2)))
	(make-instance 'trivial-analysis
		       :args-used (union (args-used bindings-analysis)
					 (args-used body-analysis))
		       :code e)))))

(defun analyze-locally (e l m s &optional (whole e) (force-trivial? t))
  ;; deal with "special" declarations
  (do ((exprs (cdr e) (cdr exprs))
       (specials nil)
       (l2 l)
       (s2 s))
      ((or (null exprs)
	   (atom (car exprs))
	   (not (eq (caar exprs) 'declare)))
       (flet ((shadowed? (a) (member (car a) specials)))
	 (analyze-exprs exprs
			(remove-if #'shadowed? l2)
			m
			(remove-if #'shadowed? s2)
			whole
			force-trivial?)))
    (dolist (decl (cdar exprs))
      (if (eq (car decl) 'special)
	  (mapc #'(lambda (v) (pushnew v specials))
		(cdr decl))))))

(defun analyze-macrolet (e l m s)
  ;; loses on &environment.  Eit.
  (do ((m2 m)
       (clauses (second e) (cdr clauses)))
      ((null clauses) (analyze-locally `(locally ,@(cddr e)) l m2 s e nil))
    (let* ((clause (car clauses))
	   (name (first clause))
	   (ll (second clause))
	   (body (cddr clause))
	   (qargs (gensym)))
      (push (cons name (eval `(lambda (&rest ,qargs)
				(destructuring-bind ,(cons (gensym) ll)
				    (cons ,name ,qargs)
				  ,@body))))
	    m2))))

(defun analyze-setq (e l m s)
  (labels ((every-other (l)
	     (if (and (consp l) (consp (cdr l)))
		 (cons (cadr l) (every-other (cddr l)))
	       nil)))
    (analyze-exprs (every-other (cdr e)) l m s) e))

(defun analyze-symbol-macrolet (e l m s)
  (do ((s2 s)
       (clauses (second e) (cdr clauses)))
      ((null clauses) (analyze-locally `(locally ,@(cddr e))
				       (set-difference l s2 :key #'car)
				       m s2 e nil))
    (let* ((clause (car clauses))
	   (name (first clause))
	   (expansion (second clause)))
      (push (cons name expansion) s2))))

(defun analyze-tagbody (e l m s)
  (analyze-exprs (remove-if #'atom (cdr e)) l m s e))

;;;

(defun analyze-and/or (e l m s)
  (let ((subanalyses (mapcar #'(lambda (x) (analyze-expr x l m s))
			     (cdr e))))
    (make-instance (case (car e)
		     (and 'and-analysis)
		     (or 'or-analysis))
		   :terms subanalyses
		   :args-used (reduce #'union subanalyses :key #'args-used))))

(defun analyze-eql (e l m s)
  (flet ((make-eql (ea mc)
		   (make-instance 'eql-analysis
				  :index (index-of ea)
				  :chain (chain-of ea)
				  :args-used (args-used ea)
				  :target (value-of mc))))
    (let* ((subanalyses (mapcar #'(lambda (x) (analyze-expr x l m s))
				(cdr e)))
	   (types (mapcar #'type-of subanalyses)))
      (cond ((equal types '(extraction-analysis manifest-constant))
	     (make-eql (first subanalyses) (second subanalyses)))
	    ((equal types '(manifest-constant extraction-analysis))
	     (make-eql (second subanalyses) (first subanalyses)))
	    (t (make-instance 'trivial-analysis
			      :args-used (reduce #'union subanalyses
						 :key #'args-used)
			      :code e))))))

(defun analyze-not (e l m s)
  (let ((base (analyze-expr (second e) l m s)))
    (if (typep base 'extraction-analysis)
	(make-instance 'extraction-analysis
		       :index (index-of base)
		       :chain (cons #'not (chain-of base))
		       :args-used (args-used base))
      (make-instance 'not-analysis
		     :base base))))

(defun analyze-typep (e l m s)
  (let ((subanalyses (mapcar #'(lambda (x) (analyze-expr x l m s))
			     (cdr e))))
    (if (and (typep (first subanalyses) 'extraction-analysis)
	     (typep (second subanalyses) 'manifest-constant))
	(make-instance 'typecheck-analysis
		       :index (index-of (first subanalyses))
		       :chain (chain-of (first subanalyses))
		       :args-used (args-used (first subanalyses))
		       :target (value-of (second subanalyses)))
      (make-instance 'trivial-analysis
		     :args-used (reduce #'union subanalyses
					:key #'args-used)
		     :code e))))

(defconstant *analysis-helpers*
  ;; omitted (potentially evaluate any "argument"):
  ;;   catch, if, multiple-value-{call,prog1}, progn, progv, throw,
  ;;   unwind-protect
  ;; quote is also omitted because constantp takes care of it.
  `((block . ,#'analyze-block)
    (eval-when . ,#'analyze-eval-when)
    (flet . ,#'analyze-flet/labels)
    (function . ,#'analyze-function)
    (go . ,#'(lambda (e l m s) (declare (ignore l m s)) (null-triv e)))
    (labels . ,#'analyze-flet/labels)
    (lambda . ,#'analyze-lambda)
    (let . ,#'analyze-let/let*)
    (let* . ,#'analyze-let/let*)
    (load-time-value . ,#'analyze-first-arg)
    (locally . ,#'analyze-locally)
    (macrolet . ,#'analyze-macrolet)
    (return-from . ,#'analyze-second-arg)
    (setq . ,#'analyze-setq)
    (symbol-macrolet . ,#'analyze-symbol-macrolet)
    (tagbody . ,#'analyze-tagbody)
    (the . ,#'analyze-second-arg)
    ;;
    (and . ,#'analyze-and/or)
    (eql . ,#'analyze-eql)
    (not . ,#'analyze-not)
    (or . ,#'analyze-and/or)
    (typep . ,#'analyze-typep)))

(defmethod build-pred-internal ((analysis trivial-analysis) lambda-list)
  (do ((ll (cons 'auxv lambda-list) (cdr ll))
       (used (args-used analysis))
       (code (code-of analysis))
       (index -1 (1+ index)) ;; starts at -1 to account for auxv.
       (auxv-count 0)
       rnorm ropt rest rkey ignores last-ll-key remaining-optionals smlcl)
      ((null ll) (if (and (not rest) (null rkey))
		     (setf rest (gensym)
			   ignores (cons rest ignores)))
       (make-instance 'test-predicate
		      :test (eval `(lambda (,@(reverse rnorm)
					    ,@(and ropt
						   (cons '&optional
							 (reverse ropt)))
					    ,@(and rest `(&rest ,rest))
					    ,@(and rkey `(&key
							  ,@(reverse rkey)
							  &allow-other-keys)))
				     ,@(and ignores
					    `((declare (ignore ,@ignores))))
				     ,(if smlcl
					  `(symbol-macrolet ,smlcl ,code)
					code)))))
    (let* ((carll (car ll))
	   (arg (extract-arg-name carll (eq last-ll-key '&key)))
	   (used? (member arg used))
	   (arg2 (and (consp carll) (third carll)))
	   (used2? (member arg2 used)))
      (cond ((eq carll '&allow-other-keys) nil) ; do nothing (always on)
	    ((member carll '(&optional &rest &key &aux))
	     (setf last-ll-key carll)
	     (if (eq carll '&optional)
		 (setf remaining-optionals (cdr ll))))
	    ((member carll lambda-list-keywords)
	     (error "Unsupported lambda list keyword ~S" carll))
	    ((and (null last-ll-key) (null (cdr used)) (eq arg (car used))
		  (>= index 0))
	     (return-from build-pred-internal
	       (make-instance 'projected-unary-predicate
			      :index index
			      :base (make-instance
				     'test-predicate
				     :test (eval `(lambda (,arg)
						    ,(code-of analysis)))
				     :pass-auxv nil))))
	    ((null last-ll-key) (push arg rnorm) ; don't want specializer!
				(if (not used?) (push arg ignores)))
	    ((and (eq last-ll-key '&rest) used?) (setf rest arg))
	    ((eq last-ll-key '&aux)
	     (if used? (push `(,arg (aref auxv ,auxv-count)) smlcl))
	     (incf auxv-count))
	    ((and (eq last-ll-key '&key) (or used? used2?))
	     (if used2? (progn (push carll rkey)
			       (if (not used?) (push arg ignores)))
	       (push (list arg (second carll)) rkey)))
	    ((and (eq last-ll-key '&optional) (or used? used2?))
	     (mapc #'(lambda (x) (push (car-or-identity x) ropt))
		   (ldiff remaining-optionals ll))
	     (setf remaining-optionals (cdr ll))
	     (if used2? (progn (push carll ropt)
			       (if (not used?) (push arg ignores)))
	       (push (list arg (second carll)) ropt)))))))

(defmethod build-pred-internal ((analysis manifest-constant) ll)
  (declare (ignore ll))
  (make-instance 'constant-predicate :value (value-of analysis)))

(defmethod build-pred-internal ((analysis extraction-analysis) ll)
  (declare (ignore ll))
  (let* ((chain0 (chain-of analysis))
	 (not? (eql (car chain0) #'not))
	 (chain (if not? (cdr chain0) chain0))
	 (test (make-instance 'test-predicate
			      :test (or (car chain) #'identity)
			      :pass-auxv nil))
	 (subbase (cond ((null chain) test)
			((null (cdr chain)) test)
			(t (make-instance 'extracting-unary-predicate
					  :accessors (cdr chain)
					  :base test))))
	 (base (make-instance 'projected-unary-predicate
			      :index (index-of analysis)
			      :base subbase)))
    (if not?
	(make-instance 'not-predicate :base base)
      base)))

(defmethod build-pred-internal ((analysis extraction-or-simple-test) ll)
  (declare (ignore ll))
  (let ((base (make-instance (if (typep analysis 'typecheck-analysis)
				 'typecheck-predicate
			       'equality-predicate)
			     :target (target-of analysis)))
	(chain (chain-of analysis)))
    (make-instance 'projected-unary-predicate
		   :index (index-of analysis)
		   :base (if chain
			     (make-instance 'extracting-unary-predicate
					    :accessors chain :base base)
			   base))))

(defmethod build-pred-internal ((analysis not-analysis) ll)
  (make-not (build-pred-internal (base-of analysis) ll)))

(defmethod build-pred-internal ((analysis compound-analysis) ll)
  (funcall (typecase analysis
	     (and-analysis #'make-and)
	     (or-analysis #'make-or))
	   (mapcar #'(lambda (a) (build-pred-internal a ll))
		   (terms-of analysis))))

(provide 'pc-build)
