(require 'pd-package)
(in-package :predicate-dispatch)

(require 'pc-build)

;; Known issues with this code:
;;  * It's not particularly efficient.
;;  * It assumes all method definitions appear at toplevel.
;;  * It has bogus semantics for specificity.  (Mostly fixed, but bogus
;;    behavior can still occur when code mixes defmethod and defpdmethod.)
;;  * CMUCL spews "Note: Deleting unused function NEXT-METHOD-P."

(defclass gf-with-predicate-dispatching (standard-generic-function)
  ()
  (:metaclass funcallable-standard-class)
  (:default-initargs :method-class (find-class 'predicate-dispatched-method)))

(defmethod update-instance-for-different-class :before
  (old (new gf-with-predicate-dispatching) &rest junk)
  ;; AFAICT, change-class ignores default initargs.
  (declare (ignore junk))
  (unless (slot-boundp old 'method-class)
    (setf (slot-value new 'method-class)
	  (find-class 'predicate-dispatched-method))))

(defclass predicate-dispatched-method (standard-method)
  ())


;; Turn the predicate list into a qualifier.  This approach has two
;; major advantages:
;;  * There's a portable way to pull the predicate list back out.
;;  * It doesn't yield "redefinitions" with identical specializers and
;;    qualifiers but different predicates.

(defmacro defpdmethod (name lambda-list pred-list &rest body)
  (ensure-generic-function name
			   :generic-function-class
			   'gf-with-predicate-dispatching)
  (let ((auxtail (member '&aux lambda-list))
	(pred (build-predicate lambda-list pred-list)))
    (if auxtail
	(do ((auxvars (cdr auxtail) (cdr auxvars))
	     (index 0 (1+ index))
	     (let-clauses nil))
	    ((null auxvars)
	     `(defmethod ,name ,pred ,(ldiff lambda-list auxtail)
		(let ,let-clauses ,@body)))
	  (push `(,(car-or-identity (car auxvars)) (aref auxv ,index))
		let-clauses))
      `(defmethod ,name ,pred ,lambda-list ,@body))))


(defmethod predicate-of ((method standard-method))
  (declare (ignore method))
  *true*)

(defmethod pq-of ((method predicate-dispatched-method))
  (find-if #'predicate-qualifier? (method-qualifiers method)))

(defmethod predicate-of ((method predicate-dispatched-method))
  (predicate-of (pq-of method)))

(defun sort-methods (methods)
  (stable-sort (copy-list methods) #'implies? :key #'predicate-of))


(defmethod compute-applicable-methods-using-classes
  ((gf gf-with-predicate-dispatching) classes)
  (multiple-value-bind (methods memoizable)
      (call-next-method)
    (values (sort-methods methods) memoizable)))

(defmethod compute-applicable-methods ((gf gf-with-predicate-dispatching) args)
  (sort-methods (call-next-method)))


(defmethod make-method-lambda ((gf gf-with-predicate-dispatching)
                               (method predicate-dispatched-method)
                               lambda-expression
                               environment)
  `(lambda (args remaining-methods &optional auxv)
     (,(call-next-method gf method
	 `(lambda (&rest args)
	    (let ((next-auxv nil))
	      (labels ((find-next-method ()
		         (if remaining-methods
			     (multiple-value-bind (applies? av)
				 (let ((pq (pq-of (car remaining-methods))))
				   (evaluate-predicate
				    (predicate-of pq) args
				    (make-array (auxv-count-of pq)
						:initial-element nil)))
			       (cond (applies? (setf next-auxv av)
					       t)
				     (t (pop remaining-methods)
					(find-next-method))))
			   nil)))
		(cond (auxv
		       (apply #'(lambda ,(cadr lambda-expression)
				  (labels ((next-method-p ()
				             (or next-auxv (find-next-method)))
					   (call-next-method (&rest cnm-args)
				             (unless (next-method-p)
					       (error "No next method for ~A."
						      ',(generic-function-name
							 gf)))
					     (funcall (method-function
						       (car remaining-methods))
						      (or cnm-args args)
						      (cdr remaining-methods)
						      next-auxv)))
				    ,@(cddr lambda-expression)))
			      args))
		      ((find-next-method)
		       (funcall (method-function (car remaining-methods))
				args (cdr remaining-methods) next-auxv))
		      (t (error "No applicable method for ~A on ~S."
				',(generic-function-name gf) args))))))
	 environment)
      args (cdr remaining-methods))))

(defmethod compute-effective-method ((gf gf-with-predicate-dispatching)
                                     method-combination
                                     methods)
  (declare (ignore method-combination))
  `(call-method ,(car methods) ,methods))

(provide 'predicate-dispatch)
