(in-package :weyl)
(require 'predicate-dispatch)
(use-package 'predicate-dispatch)

;; This code assumes that the Bessel and gamma functions have been defined.

(defpdmethod struve-h (r z)
  (let ((d (or (first-domain r z) *general*)))
    (make-ge-funct d (make-function d 'struve-h 2) r z)))

(defpdmethod struve-h ((r ratio) z)
  ((> r 0)
   (not (zerop z))
   (= (denominator r) 2))
  ;; Weyl doesn't support symbolic sums. :-/
  (do ((m 0 (1+ m))
       (sum 0 (+ sum (/ (* (gamma (+ m 1/2))
			   (expt (/ z 2) (+ (* -2 m) r -1)))
			(gamma (+ r 1/2 (* -1 m)))))))
      ((> m r) (+ (bessel-y r z) (/ sum pi)))))

(defpdmethod struve-h ((r ratio) z)
  ((< r 0)
   (not (zerop z))
   (= (denominator r) 2))
  (* (expt -1 (- -1/2 r)) (bessel-j (- r) z)))

(defpdmethod struve-h (nu z)
  ((zerop z))
  0)

(defpdmethod struve-h ((nu number) (z number))
  ((or (floatp nu) (floatp z)))
  ;; don't give inexact results for exact inputs.
  (do ((s 0 (+ s (/ zf g1 g2)))
       (so -1 s)
       (z2 (- (expt (/ z 2) 2)))
       (k1 3/2 (1+ k1))
       (k2 (+ nu 3/2) (1+ k2))
       (g1 (gamma 3/2) (* g1 k1))
       (g2 (gamma (+ nu 3/2)) (* g2 k2))
       (zf ((expt (/ z 2)) (+ nu 1)) (* zf z2)))
      ((= s so) s))
  )

;; not sure how to work derivative or print form in, given issues with
;; mixing defmethod and defpdmethod. :-/ (Subclassing ge-application
;; also seems to be out, given Weyl's treatment of it.)
