(in-package :weyli)
(require 'predicate-dispatch)
(use-package 'predicate-dispatch)

(defun tr-sin? (expr)
  (and (typep expr 'ge-application)
       (equal (name-of (funct-of expr)) "sin")))

(defun tr-cos? (expr)
  (and (typep expr 'ge-application)
       (equal (name-of (funct-of expr)) "cos")))

(defun tr-arg-of (test terms)
  (first (args-of (find-if test terms))))

(export 'trig-linearize)

(defpdmethod trig-linearize ((expr general-expression))
  ()
  expr)

(defpdmethod trig-linearize ((expr list))
  (mapcar #'trig-linearize expr))

(defpdmethod trig-linearize ((expr ge-plus))
  (make-ge-plus (domain-of expr)
		(mapcar #'trig-linearize (terms-of expr))))

(defpdmethod trig-linearize ((expr ge-times) &aux (terms (terms-of expr)))
  ((member-if #'tr-sin? terms)
   (member-if #'tr-cos? terms))
  (let ((x (tr-arg-of #'tr-sin? terms))
	(y (tr-arg-of #'tr-cos? terms))
	(other-terms (remove-if #'tr-sin?
				(remove-if #'tr-cos? terms :count 1)
				:count 1)))
    (make-ge-times (domain-of expr)
		   (list* 1/2 (+ (sin (+ x y)) (sin (- x y)))
			  (trig-linearize other-terms)))))

(defpdmethod trig-linearize ((expr ge-times) &aux (terms (terms-of expr)))
  ((>= (count #'tr-sin? terms) 2))
  (let* ((px (position-if #'tr-sin? terms))
	 (x (first (args-of (elt terms px))))
	 (y (first (args-of (find-if #'tr-sin? terms :start (1+ px)))))
	 (other-terms (remove-if #'tr-sin? terms :count 2)))
    (make-ge-times (domain-of expr)
		   (list* 1/2 (- (cos (- x y)) (cos (+ x y)))
			  (trig-linearize other-terms)))))

(defpdmethod trig-linearize ((expr ge-times) &aux (terms (terms-of expr)))
  ((>= (count #'tr-cos? terms) 2))
  (let* ((px (position-if #'tr-cos? terms))
	 (x (first (args-of (elt terms px))))
	 (y (first (args-of (find-if #'tr-cos? terms :start (1+ px)))))
	 (other-terms (remove-if #'tr-cos? terms :count 2)))
    (make-ge-times (domain-of expr)
		   (list* 1/2 (+ (cos (- x y)) (cos (+ x y)))
			  (trig-linearize other-terms)))))

(provide 'trig-rearrangement)
